﻿using ProjectGasMachine.Drawnings;
using ProjectGasMachine.Exceptions;

namespace ProjectGasMachine.CollectionGenericObjects;

/// <summary>
/// класс-хранилище
/// </summary>
/// <typeparam name="T"></typeparam>
public class StorageCollection<T>
    where T : DrawningMachine
{
    /// <summary>
    /// словарь (хранилище) с коллекциями
    /// </summary>
    private Dictionary<CollectionInfo, ICollectionGenericObjects<T>> _storages;

    /// <summary>
    /// Возвращение списка названий коллекций
    /// </summary>
    public List<CollectionInfo> Keys => _storages.Keys.ToList();

    /// <summary>
    /// Ключевое слово, с которого должен начинаться файл
    /// </summary>
    private readonly string _collectionKey = "CollectionsStorage";
    /// <summary>
    /// Разделитель для записи ключа и значения элемента словаря
    /// </summary>
    private readonly string _separatorForKeyValue = "|";
    /// <summary>
    /// Разделитель для записей коллекции данных в файл
    /// </summary>
    private readonly string _separatorItems = ";";

    /// <summary>
    /// конструктор
    /// </summary>
    public StorageCollection() 
    {
        _storages = new Dictionary<CollectionInfo, ICollectionGenericObjects<T>>();
    }

    /// <summary>
    /// добваление коллекции в хранилище
    /// </summary>
    /// <param name="name">название коллекции</param>
    /// <param name="collectionType">тип коллекции</param>
    public void AddCollection(CollectionInfo info) 
    {
        if (info == null || _storages.ContainsKey(info)) 
        {
            return;
        }

        if (info.CollectionType == CollectionType.Massive)
        {
            _storages.Add(info, new MassiveGenericObjects<T>());
        }

        if (info.CollectionType == CollectionType.List)
        {
            _storages.Add(info, new ListGenericObjects<T>());
        }
    }

    /// <summary>
    /// Удаление коллекции
    /// </summary>
    /// <param name="name"></param>
    public void DelCollection(CollectionInfo info) 
    {
        if (info == null || !_storages.ContainsKey(info)) 
        {
            return;
        }

        _storages.Remove(info);
    }

    /// <summary>
    /// Доступ к коллекции
    /// </summary>
    /// <param name="name">Название коллекции</param>
    /// <returns></returns>
    public ICollectionGenericObjects<T>? this[CollectionInfo info] 
    {
        get 
        {
            if (_storages.ContainsKey(info)) 
            {
                return _storages[info];
            }

            return null;
        }
    }


    /// <summary>
    /// Сохранение информации по автомобилям в хранилище в файл
    /// </summary>
    /// <param name="filename">Путь и имя файла</param>
    public void SaveData(string filename)
    {
        if (_storages.Count == 0)
        {
            throw new NullReferenceException("В хранилище отсутствуют коллекции для сохранения");
        }

        if (File.Exists(filename))
        {
            File.Delete(filename);
        }

        using (StreamWriter sw = new StreamWriter(filename))
        {
            sw.Write(_collectionKey);

            foreach (KeyValuePair<CollectionInfo, ICollectionGenericObjects<T>> value in _storages)
            {
                sw.Write(Environment.NewLine);
                // не сохраняем пустые коллекции
                if (value.Value.Count == 0)
                {
                    continue;
                }

                sw.Write(value.Key);
                sw.Write(_separatorForKeyValue);
                sw.Write(value.Value.MaxCount);
                sw.Write(_separatorForKeyValue);

                foreach (T? item in value.Value.GetItems())
                {
                    string data = item?.GetDataForSave() ?? string.Empty;

                    if (string.IsNullOrEmpty(data))
                    {
                        continue;
                    }
                    sw.Write(data);
                    sw.Write(_separatorItems);
                }
            }
        }

    }

    /// <summary>
    /// Загрузка информации по автомобилям в хранилище из файла
    /// </summary>
    /// <param name="filename">Путь и имя файла</param>
    public void LoadData(string filename)
    {
        if (!File.Exists(filename))
        {
            throw new FileNotFoundException("Файл не существует");
        }

        using (StreamReader sr = new(filename))
        {
            string line = sr.ReadLine();

            if (line == null || line.Length == 0)
            {
                throw new FileFormatException("В файле нет данных");
            }
            if (!line.Equals(_collectionKey))
            {
                //если нет такой записи, то это не те данные
                throw new FileFormatException("В файле неверные данные");
            }
            _storages.Clear();

            while ((line = sr.ReadLine()) != null)
            {
                string[] record = line.Split(_separatorForKeyValue, StringSplitOptions.RemoveEmptyEntries);
                if (record.Length != 3)
                {
                    continue;
                }

                CollectionInfo? collectionInfo = CollectionInfo.GetCollectionInfo(record[0]) ??
                throw new Exception("Не удалось определить информацию коллекции: " + record[0]);
                ICollectionGenericObjects<T>? collection = StorageCollection<T>.CreateCollection(collectionInfo.CollectionType) ??
                    throw new Exception("Не удалось создать коллекцию");
                collection.MaxCount = Convert.ToInt32(record[1]);

                string[] set = record[2].Split(_separatorItems, StringSplitOptions.RemoveEmptyEntries);
                foreach (string elem in set)
                {
                    if (elem?.CreateDrawningMachine() is T machine)
                    {
                        try 
                        {
                            if (collection.Insert(machine, new DrawningMachineEqutables()) == -1)
                            {
                                throw new InvalidOperationException("Объект не удалось добавить в коллекцию: " + record[3]);
                            }
                        }
                        catch (CollectionOverflowException ex)
                        {
                            throw new OverflowException("Коллекция переполнена", ex);
                        }
                        catch (ObjectAlreadyInCollectionException ex)
                        {
                            throw new InvalidOperationException("Объект уже присутствует в коллекции", ex);
                        }
                    }
                }
                _storages.Add(collectionInfo, collection);
            }
        }
    }

    /// <summary>
    /// Создание коллекции по типу
    /// </summary>
    /// <param name="collectionType"></param>
    /// <returns></returns>
    private static ICollectionGenericObjects<T>? CreateCollection(CollectionType collectionType)
    {
        return collectionType switch
        {
            CollectionType.Massive => new MassiveGenericObjects<T>(),
            CollectionType.List => new ListGenericObjects<T>(),
            _ => null,
        };
    }

}
